#!/usr/bin/perl 

eval 'exec /usr/bin/perl  -S $0 ${1+"$@"}'
    if 0; # not running under some shell

# See embedded POD below for further information

# TODO
# - Add more error messages and checks

use strict;
use warnings;

#DEBUG
use Data::Dumper;
#DEBUG

use FindBin;
use lib "$FindBin::RealBin/../lib";

use locale;
use POSIX qw(setsid);
# DateTime not needed by now, but this would be necessary, because Hades.pm is
# NOT loaded.
#use DateTime;
#use DateTime::Locale;
#BEGIN {
#  if (DateTime::Locale->load(setlocale(LC_TIME))) {
#    DateTime->DefaultLocale(setlocale(LC_TIME));
#  }
#}

use Log::Dispatch;
use Log::Dispatch::File;
use Log::Dispatch::Syslog;
use Log::Dispatch::Screen;
use Getopt::Long 2.32 qw(:config auto_help auto_version prefix=--);
  # TODO prefix=-- necessary to allow "-" ! Bug??
use Pod::Usage;
use Config::General;

use Socket;

use File::Spec;

use IPC::Shareable;
use HTTP::Response;
use HTTP::Daemon;
use HTTP::Daemon::SSL qw(debug4);

use NMWG;
use perfSONAR;
use perfSONAR::Echo;
use perfSONAR::SOAP::Message;
use perfSONAR::Client::LS;

use vars qw($VERSION);
$VERSION     = 0.51;

my $pidfile_ok = 0; # important for END block


#
# Parse Configuration (commandline and file)
#

my (
  $configfile, $noconfig,
  $detach, $syslog, $logfile, $nologfile, $pidfile, $nopidfile,
  $loglevel, $verbose, $syslog_host, $syslog_ident, $syslog_facility,
  $ls_register, $keepalive, $max_proc,
  @ls_url, $auth, $as_url,
  $hostname, $port, $organization, $contact,
  $ssl, $nossl,
);

my %services;
my %messages;
my %lsKeys;

GetOptions(
  "config=s"      => \$configfile,
  "noconfig"      => \$noconfig,
  "detach|D!"     => \$detach,
  "logfile:s"     => \$logfile,
  "nologfile"     => \$nologfile,
  "pidfile:s"     => \$pidfile,
  "nopidfile"     => \$nopidfile,
  "syslog!"       => \$syslog,
  "syslog-host=s"     => \$syslog_host,
  "syslog-ident=s"    => \$syslog_ident,
  "syslog-facility=s" => \$syslog_facility,
  "loglevel=s"    => \$loglevel,
  "verbose|v"     => \$verbose,
  "register!"     => \$ls_register,
  "keepalive=i"   => \$keepalive,
  "max_proc=i"    => \$max_proc,
  "port=s"        => \$port,
  "auth!"         => \$auth,
  "as_url=s"      => \$as_url,
  "ssl"	          => \$ssl,
  "nossl"         => \$nossl,
) or pod2usage(2);

# Determine and load config file
my %Config = ();
my $Config;
if ($noconfig) {
  $configfile = undef;
} else {
  $configfile ||= "$FindBin::RealBin/../etc/oppd.conf";
  $Config = Config::General->new(
    -ConfigFile => $configfile,
    -ApacheCompatible => 1,
    -AutoTrue => 1,   # Could bring in some trouble, but it is really nice ;)
    -CComments => 0,  # Parsing is obviously broken in 2.36!
                      # Comments are found everywhere...
  );
  %Config = $Config->getall;
}

# We already need the Log::Dispatch object for option verification
my $log = Log::Dispatch->new();

#
# Calculate options
# First not "undef" value is used.
# Order: command line, config file, default
#
$detach     = get_opt($detach,    $Config{detach},  1);
$nologfile  = get_opt($nologfile, 0); # No nologfile entry in config file!
if ($nologfile) {
  $logfile = undef;
} else {
  $logfile = get_opt($logfile,    $Config{logfile}, 0);
  if (!$logfile && $logfile ne "") {
    # logfile disabled
    $logfile = undef;
  } elsif ($logfile eq "1" || $logfile eq "") {
    # logfile enabled in configuration file or via --logfile without value
    $logfile = "/var/log/oppd.log";
  }
}
$nopidfile  = get_opt($nopidfile, 0); # No nopidfile entry in config file!
if ($nopidfile) {
  $pidfile = undef;
} else {
  $pidfile = get_opt($pidfile,    $Config{pidfile}, 1);
  if (!$pidfile && $pidfile ne "") {
    # pidfile disabled
    $pidfile = undef;
  } elsif ($pidfile eq "1" || $pidfile eq "") {
    # pidfile enabled in configuration file or via --pidfile without value
    $pidfile = "/var/run/oppd.pid";
  }
}
$syslog       = get_opt($syslog,      $Config{syslog},      0);
$syslog_host  = get_opt($syslog_host, $Config{'syslog-host'}, "");
$syslog_ident =
  get_opt($syslog_ident,    $Config{'syslog-ident'},    "oppd");
$syslog_facility =
  get_opt($syslog_facility, $Config{'syslog-facility'}, "daemon");
$loglevel     = get_opt($loglevel,    $Config{loglevel},    "notice");
$verbose      = get_opt($verbose,     0); # No verbose entry in config file!
if ($verbose) {
  $loglevel = "info";
} else {
  pod2usage( { -message => "Invalid log level: $loglevel",
               -exitval => 2 } )  unless $log->level_is_valid($loglevel);
}
#$ls_url     = get_opt($ls_url,    $Config{ls_url},    "");
#LS URL can be one or more, take care of that:
my $ls_opt = $Config{ls_url};
if(ref($ls_opt) eq "ARRAY"){
  @ls_url = @{$ls_opt};
} else {
  @ls_url = ($ls_opt);
}
$ls_register = get_opt($ls_register,  $Config{ls_register},   0);
$keepalive  = get_opt($keepalive, $Config{keepalive},  3600);
$max_proc   = get_opt($max_proc,  $Config{max_poc},   5);
$hostname   = get_opt($hostname,  $Config{hostname},   0);
$port       = get_opt($port,      $Config{port},   8090);
$organization = get_opt($organization, $Config{organization},  0);
$contact    = get_opt($contact,   $Config{contact},   0);
$auth       = get_opt($auth,      $Config{auth},    0);
$as_url     = get_opt($as_url,    $Config{as_url},  "");
$ssl        = get_opt($ssl,       $Config{ssl}, 0);
$nossl      = get_opt($nossl,     $Config{nossl}, 0);
$ssl = 0 if $nossl;

%services = %{$Config{service}}
  or pod2usage(
    "No services specified, that the perfSONAR daemon should provide (see config file)"
  );


#
# Start logging ($log already initialised above)
#

$perfSONAR::log = $log;

if (defined $logfile) {

  #check if $logfile is an absolute path, and add current path if not
  if (!File::Spec->file_name_is_absolute($logfile)){
    $logfile = File::Spec->rel2abs($logfile);
  }

  eval {
    use locale;
    use POSIX qw(locale_h strftime);
    # "use locale", POSIX::strftime and localtime faster than DateTime
    # and this has quite an impact on debug loging. Detected via SmokePing:
    # http://perfsonar.acad.bg/smokeping
    #use DateTime;
    #use DateTime::Locale;
    #BEGIN {
    #  if (DateTime::Locale->load(setlocale(LC_TIME))) {
    #    DateTime->DefaultLocale(setlocale(LC_TIME));
    #  }
    #}
    $log->add(
      Log::Dispatch::File->new(
        name => 'file',
        min_level => $loglevel,
        filename => $logfile,
        mode => 'append', close_after_write => 1, autoflush => 1,
        callbacks => sub {
          my %p=@_;
          return
            #DateTime->now(time_zone=>"local")->strftime("%c") .
            strftime("%FT%T",localtime) .
            " $p{level} $p{message}\n";
        },
      )
    );
  };
  die "Cannot write to log file '$logfile': $@\n" if $@;
}
if ($syslog) {
  eval {
    my $syslog_socket = 'unix';
    if ($syslog_host) {
      $Sys::Syslog::host = $syslog_host;
      $syslog_socket = 'inet';
    }
    $log->add(
      Log::Dispatch::Syslog->new(
        name => 'syslog',
        min_level => $loglevel,
        ident => "$syslog_ident",
        facility => "$syslog_facility",
        socket => "$syslog_socket",
        logopt => "ndelay",
        # callback for SD stuff
        #TODO Not nice! How can we change this?
        callbacks => sub {
          my %p=@_;
          if ($p{service}) {
            $p{service} =~ s/\//\_/g;
            $p{service} .= ".";
          } else {
            $p{service} = "";
          }
          return "OPPD." . $p{service} . uc($p{level}) . "% $p{message}\n";
        },
        #mode => 'append', close_after_write => 0, autoflush => 1,
      )
    );
  };
  die "Cannot write to syslog: $@\n" if $@;
}
unless ($detach) {
  $log->add(
    Log::Dispatch::Screen->new(
      name => 'screen',
      min_level => $loglevel,
      stderr    => 1,
      callbacks => sub {
        my %p=@_; return "$p{level}: $p{message}\n";
      },
    )
  );
}

# More flexible die:
# Put error into Log and afterwards die with same message.
# Also handy, because in the following code $@ is undef in die call:
# $log->error($@); die $@;
$SIG{__DIE__} = sub {
  die @_ if $^S; # Ignore dies from evals
  my $logmsg = join " - ", @_;
  chomp $logmsg; # No new line for Log::Dispatch !
  # We should only be called with initialised $log, but we can be a bit
  # more friendly by only using it if it was initialised:
  $log->error($logmsg) if defined $log && UNIVERSAL::isa($log,'Log::Dispatch');
  die @_;
};

# More flexible warn:
# Put error into Log and afterwards warn with same message.
$SIG{__WARN__} = sub {
  my $logmsg = join " - ", @_;
  chomp $logmsg; # No new line for Log::Dispatch !
  # We should only be called with initialised $log, but we can be a bit
  # more friendly by only using it if it was initialised:
  $log->warning($logmsg)
    if defined $log && UNIVERSAL::isa($log,'Log::Dispatch');
  warn @_;
};


#
# Load Authentication module for AA
#

if ($auth){
  if ($as_url eq ""){
    die ("Authentication not possible: variable $as_url not set!\n");
  }
  eval require perfSONAR::Auth;
  if ($@){
    die ("Error loading module perfSONAR::Auth: $@\n");
  }
}

#
# Load data modules for services
#

foreach my $service (keys %services){
  my $module = $services{$service}->{module};
  eval "require perfSONAR::$module";
  if ($@){
    die ("Error loading module perfSONAR::$module: $@\n");
  }
  $services{$service}->{handler} =
    "perfSONAR::$module"->new(%{$services{$service}->{module_param}});
}

# Set modules in dispatch module/class:
%perfSONAR::services = %services;

#
# Daemonize
#

#First check pidfile path to be absolute!
if ($pidfile){
  if (!File::Spec->file_name_is_absolute($pidfile)){
    $pidfile = File::Spec->rel2abs($pidfile);
  }
}

if ($detach) {
  # Fork once, and let the parent exit.
  my $pid = fork;
  exit if $pid;
  defined($pid) or die("Could not fork: $!\n");

  # Dissociate from the controlling terminal that started us and stop being
  # part of whatever process group we had been a member of.
  setsid() or die("Cannot start a new session: $!\n");

  # Forks another child process and exits first child. This prevents the
  # potential of acquiring a controlling terminal
  $pid = fork;
  exit if $pid;
  defined($pid) or die("Could not fork: $!\n");

  # chdir and set umask
  chdir '/' or die("Cannot chdir to '/': $!\n");
  #umask 0;

  setup_pidfile() if defined $pidfile;
    # Do it before closing file handles! We need the error messages!

  # Close default file handles
  close STDIN or die("Could not close STDIN: $!\n");
  close STDOUT or die("Could not close STDOUT: $!\n");
  close STDERR or die("Could not close STDERR: $!\n");
  # Reopen stderr, stdout, stdin to /dev/null
  open(STDIN,  "</dev/null");
  open(STDOUT, ">/dev/null");
  open(STDERR, ">/dev/null");
} else {
  setup_pidfile() if defined $pidfile;
}

# die on typical signals
my $time_to_die = 0;
my $time_to_reset = 0;
$SIG{INT} = $SIG{TERM} = sub {
  $log->notice("Signal caught - exiting");
  exit 1;
};
$SIG{HUP} = sub {
  $log->notice("SIGHUP caught - NO RELOAD SUPPORTED AT THE MOMENT");
  #TODO
};
$SIG{PIPE} = 'IGNORE';
$SIG{CHLD} = 'IGNORE'; #avoid zombie children


#
# Inform that everything looks good
#

$log->notice("oppd service started");
$log->info("available services: " . join(",", sort keys(%services)));
$log->info("PID $$ written to $pidfile") if defined $pidfile;

#
# Start "daemon", the network side of the job ;-)
# 

#TODO: enable tracing output for our own SOAP implementation
#      Use Log::Dispatcher for debug/error handling somehow?

my $http_daemon;

if ($ssl){
  $http_daemon = HTTP::Daemon::SSL->new(
    LocalPort => $port, ReuseAddr => 1, SSL_verify_mode => 0x02
  );
} else {
  $http_daemon = HTTP::Daemon->new(
    LocalPort => $port, ReuseAddr => 1
  );
}


if (!$http_daemon){
  $log->error("Stopping oppd.pl: $!");
  die "Stopping oppd.pl: $!\n";
}

#
# start process for LS registration and keepalive
#

if ($ls_register){
  if (!@ls_url){
    $log->notice("No URL for LS registration given, going on without registration!");
  } else {
    my $pid = fork();
    defined($pid) or $log->warning("Could not fork registration process: $!"); #go on without registration!
    if ($pid == 0) { #child process, start registration process
      %perfSONAR::Client::LS::services = %services;
      $perfSONAR::Client::LS::hostname = $hostname;
      $perfSONAR::Client::LS::port = $port;
      $perfSONAR::Client::LS::organization = $organization;
      $perfSONAR::Client::LS::contact = $contact;
      $perfSONAR::Client::LS::keepalive = $keepalive;
      @perfSONAR::Client::LS::ls_url = @ls_url;
      $perfSONAR::Client::LS::log = $log;

      $SIG{INT} = $SIG{TERM} = sub {
        $log->notice("Signal caught - exiting heartbeat process, deregistering services");
        deregister();
        exit 1;
      };
      heartbeat($keepalive);
    }
  }
}


#
# start process for selftest functionality
#

#selftesting functions are defined in plugin modules and
#can be configured in oppd.conf per service
=cut
my @testservices;

foreach my $service (keys %services){
  if ($services{$service}{"selftest"}){
    push @testservices, $service;
  } else {
    $log->log(
      level => "notice", service => $service,
      message => "No selftests specified for $service."
    );
  }
}

while (my $service = pop @testservices){ 
  foreach my $test (keys %{$services{$service}{"selftest"}}){      
    my $pid = fork();
    defined($pid) or $log->log(
      level => "warning", service => $service,
      message => "Could not fork selftest process $test for module $service: $!"
    );
    #last if ($pid!=0);
    if ($pid==0){ #child process, do selftest
      $SIG{INT} = $SIG{TERM} = sub {
        $log->log(
          level => "notice", service => $service,
          message => "Signal caught - exiting selftest process $test for $service"
        );
        exit 1;
      };
      while (1){
        my $result = $services{$service}->{handler}->selftest($test);
        $log->log(
          level => "notice", service => $service,
          message => "selftesting service $service: $test returned $result"
        );
        sleep($services{$service}{"selftest"}{$test});
      }
    }
  }
}
=cut

# Accept connections

#forking one process per request
my $count;
my $count_handle = tie $count, 'IPC::Shareable', undef, { destroy => 1 };
$count = 1;
while (my $c = $http_daemon->accept) {
  my $pid = fork();
  # We are going to close the new connection on one of two conditions
  #  1. The fork failed ($pid is undefined)
  #  2. We are the parent ($pid != 0)
  unless( defined $pid && $pid == 0 ) {
    $c->close;
    next;
  }
  # From this point on, we are the child.

  $http_daemon->close;  # Close the listening socket (always done in children)

  # Handle requests as they come in
  while (my $r = $c->get_request) {
    if ($count > $max_proc){ #more requests than allowed number of processes
      $log->notice("More requests than allowed number of processes, refusing service");
      my $code = 503; #RC_SERVICE_NOT_AVAILABLE
      my $r = HTTP::Response->new($code);
      my $cooo = $r->code;
      $c->send_response($r);
    } else {
      $count_handle->shlock();
      $count++;
      $count_handle->shunlock();

      my $response = new HTTP::Response;
      eval {
        my $soap_message = perfSONAR::SOAP::Message->from_http_request($r);
  
        my $nmwg_message = NMWG::Message->new( ($soap_message->body)[0] );
        #TODO: Auth
        if ($auth){
          perfSONAR::Auth::authenticate($soap_message, $nmwg_message, $as_url);
        }
        my $nmwg_response = perfSONAR->handle_request(
          $soap_message->uri, $nmwg_message
        );
        #TODO $nmwg_message <-> $nmwg_response? clone?
        #TODO what about header?
        $soap_message->body($nmwg_response->as_dom()->documentElement);
        $response->content($soap_message->as_string);
      };
      if ($@) {
        $response->content(
          perfSONAR::SOAP::Message->new(
            fault => perfSONAR::SOAP::Fault_v1_1->new(
              "", "Server.Internal", $@
            )
          )->as_string
        );
      }
      $c->send_response($response);

      $count_handle->shlock();
      $count--;
      $count_handle->shunlock();
   }
  }
  $c->close;
  #print "process id: $$\n";
}

exit 0;


### END OF MAIN ###



# Returns the first found parameter with a "defined" value
sub get_opt {
  foreach (@_) {
    return $_ if defined;
  }
  return undef;
}


END { # Always clean up PID files
  unlink $pidfile if $pidfile_ok && -e $pidfile;
}

#
# setup pid file
#
sub setup_pidfile {
  die("PID file ($pidfile) contains pid! Already running?\n")
    if -e $pidfile && -s $pidfile;
  open(PIDFILE, ">$pidfile")
    or die("Could not write PID file ($pidfile): $!\n");
  print PIDFILE "$$\n";
  $pidfile_ok = 1;
  close PIDFILE
    or die("Could not write PID file ($pidfile): $!\n");
}



__END__



=head1 NAME

B<oppd.pl> - open perl perfSONAR daemon

=head1 SYNOPSIS

B<oppd.pl> [OPTIONS]



=head1 DESCRIPTION

This is the perl perfSONAR deamon script, running different perl perfSONAR
services. 
Services can be configured in the oppd.conf file.
For more information about perfSONAR, see L<http://www.perfsonar.net/>.


=head1 OPTIONS

This is a full list of available command line options. Please keep in mind
that this script does NOT provide the normal Hades command line options
or configuration file options!
Some options might even look familiar, although they are used slightly
different!

Nearly all options have a built in default that can be overwritten using
command line arguments or variables in the configuration file. 
Arguments have precedence over variables in the configuration file.


=over


=item B<--help>

Prints a help message and exits.


=item B<--config>=F<CONFIGFILE>

Read configuration file F<CONFIGFILE> for options.

Default: F<[INSTALL-ROOT]/etc/oppd.conf>


=item B<--noconfig>

Do not read any configuration file. The parameter B<--config> is also ignored!

Default: off


=item B<--[no]detach>

Detach from terminal, aka run in background (instead of foreground).
Log messages will not be sent to F<STDERR>.

Default: on

Configuration file: B<detach>


=item B<--logfile>[=F<LOGFILE>]

Append messages to file F<LOGFILE>.

Just use B<--logfile> without the optional value to enable logging to default
log file F</var/log/oppd.log>.

You can use this option together with B<--syslog>.
Messages will then be written to both, log file and system log.

Default: off

Configuration file: B<logfile>


=item B<--nologfile>

Do not write to any log file. The parameter B<--logfile> is also ignored!

Default: off

Configuration file: use B<logfile>


=item B<--[no]syslog>

Whether messages should be written to system log.

You can use this option together with B<--logfile>.
Messages will then be written to both, log file and system log.

Default: off

Configuration file: B<syslog>


=item B<--syslog-host>=I<HOST>

Use I<HOST> as host to which system log messages are forwarded.

If this option is set to a dns name or ip address, all system log messages
are forwarded to the specified remote host.
If set to the empty string ("") logging is done locally.

Default: log locally

Configuration file: B<syslog-host>


=item B<--syslog-ident>=I<IDENT>

The string I<IDENT> will be prepended to all messages in the system log.

Default: I<oppd>

Configuration file: B<syslog-ident>


=item B<--syslog-facility>=I<FACILITY>

Use I<FACILITY> as type of program for system logging.

This string will be used as the system log facility for messages sent to
the system log.

See your C<syslog(3)> documentation for the facilities available on your
system.
Typical facilities are I<auth>, I<authpriv>, I<cron>, I<daemon>, I<kern>,
I<local0> through I<local7>, I<mail>, I<news>, I<syslog>, I<user>, I<uucp>.

Default: I<daemon>

Configuration file: B<syslog-facility>


=item B<--loglevel>=I<LOGLEVEL>

Use I<LOGLEVEL> as log level used for logging to syslog and to the log files.

This option is used for setting the verbosity of the running daemon.
The log levels available are the log levels defined by Log::Dispatch.

This is a list of values that should be accepted:
  0 = debug
  1 = info
  2 = notice
  3 = warning
  4 = err     = error
  5 = crit    = critical
  6 = alert
  7 = emerg   = emergency

Default: I<notice>

Configuration file: B<loglevel>


=item B<--verbose>

Just a handy abbreviation for B<--loglevel>=I<info>.

Default: not set, see B<--loglevel>

Configuration file: use B<loglevel>=I<info>


=item B<--pidfile>[=F<PIDFILE>]

Use PIDFILE as name of pid file.
The pid file contains the Process ID of the running oppd service.

Just use B<--pidfile> without the optional value to use the default pid file
F</var/run/oppd.pid>.

Default: F</var/run/oppd.pid>

Configuration file: B<pidfile>


=item B<--nopidfile>

Do not use a pid file. The parameter B<--pidfile> is also ignored!

Default: off

Configuration file: use B<pidfile>


=item B<--max_proc>=F<number of processes>

Maximum Number of processes to get forked for listening to requests.
This means that the service is able to handle B<--max_proc> 
numbers of requests simultaneously.

Default: 5 


=item B<--port>=PORT

Port number the service is listening to incoming requests.

Default: 8090


=item B<--[no]auth>

Use authentication with an perfSONAR Authentication Server. 
Note that you have to provide the option to specify the URL 
for the Authentication Server also.

Default: disabled


=item B<--as_url>=F<AS URL>

Provide URL for Authentication Server (see above). 
Only needed when authentication is enabled. 
Otherwise this option has no effect.

Default: none


=item B<--[no]register>

Registration to perfSONAR Lookup Server. 
When enabled, sends registration messages to Lookup Server specified by ls_url.

Default: disabled


=item B<--keepalive>=F<keepalive interval>

Interval in seconds when keepalive messages are sent to Lookup Service. 
This option has no effect, if Lookup Service registration is disabled. 
Note that the interval should be as long as possible, as this reduces 
communication overhead.

Default: none. At least 12 hours recommended.


=item B<--ls_url>=F<LS URL>

Provide URL for Lookup Server (see above). 
Only needed when Lookup Service registration is enabled. 
Otherwise this option has no effect.

Default: none


=item B<--hostname>=F<HOSTNAME> 

=item B<--organization>=F<ORGANIZATION>

=item B<--contact>=F<CONATCT>

These information are sent in the registration message to the 
Lookup Service.
If registration is not enabled, these options have no effect.

Default: none


=back

=head1 SIGNALS

Most terminating signals are caught and daemon terminates gracefully.

SIGHUP: Ignored and daemon is NOT reconfigured at the moment.



=head1 EXAMPLES

oppd.pl --port=51234 --nologfile --pidfile=oppd.pid

=head1 SEE ALSO

oppd.conf



=head1 AUTHORS

DFN Labor Erlangen, win-labor@dfn.de
