package perfSONAR;

#DEBUG
use Data::Dumper;
#/DEBUG

use strict;
use warnings;

BEGIN {
  use vars qw($VERSION @ISA @EXPORT @EXPORT_OK %EXPORT_TAGS);
  use Exporter;
  #@ISA = qw(Exporter SOAP::Server::Parameters);
  @ISA = qw(Exporter);

  # set the version for version checking
  $VERSION     = 0.51;
  # if using RCS/CVS, this may be preferred
  #$VERSION = sprintf "%d.%03d", q$Revision: 1.1 $ =~ /(\d+)/g;
  @EXPORT      = qw();          # Symbols to autoexport (:DEFAULT tag)
  @EXPORT_OK   = qw(print_log %services); # Symbols to export on request
  %EXPORT_TAGS = qw();          # Define names for sets of symbols
                                # eg: TAG => [ qw(name1 name2) ],
}

use Carp;

use perfSONAR::SOAP::HTTP::Request;
use perfSONAR::SOAP::HTTP::Response;
use perfSONAR::SOAP::HTTP::UserAgent;
#use perfSONAR::Echo;
use NMWG;
use NMWG::Message;


our $log = undef;

our %services = ();

my $echo_et = "http://schemas.perfsonar.net/tools/admin/echo/2.0";

sub handle_request {
  my $class = shift; # TODO
  my ($uri,$reqmsg) = (@_);
  my $xmlmsg = $reqmsg->as_string;
  print_log("info","Received message");
  print_log("debug","Raw Message:\n$xmlmsg");
  #print_log("debug","Parsed Message:\n" . Dumper($reqmsg));
  # How to print DOM tree? Useful???

  my $service = undef;

  #if no accesspoint is given we do not know what service is meant!
  if (!$uri){
    croak "No service specified!";
  } else {
    ($service = $uri) =~ s/^.*\/services\///;
    $service =~ s/\/$//;
    my $serviceurl = $service;
    print_log("info", "Service: $service");
    if (!$services{$service}){
      croak "Service $serviceurl not known!";
    }
  }

  my @module_ets = @{$services{$service}->{"handler"}->{"supportedEventtypes"}};
  push @module_ets, $echo_et;


  my $messagetype = $reqmsg->get_message_type();
  if ($messagetype eq "ErrorResponse"){ #error from authentication!
    return $reqmsg;
  }
  if ($messagetype eq "AuthNEERequest"){ #authorization rquest to dummy AS
    require perfSONAR::AS;
    return perfSONAR::AS::dummy();
  }
  if ($messagetype eq "SetupDataRequest"){
    $reqmsg->set_message_type("SetupDataResponse");
  }
  elsif ($messagetype eq "MeasurementRequest"){
    $reqmsg->set_message_type("MeasurementResponse");
  }
  elsif ($messagetype eq "MetadataKeyRequest"){
    $reqmsg->set_message_type("MetadataKeyResponse");
  }
  elsif ($messagetype =~ /EchoRequest/){
    $reqmsg->set_message_type("EchoResponse");
  }
  else {
    my $errorstring = "Unknown messagetype: $messagetype";
    print_log("info",$errorstring);
    $reqmsg->set_message_type("ErrorResponse");
    $reqmsg->return_result_code("error.common.action_not_supported", "$errorstring", "message");
    return $reqmsg;
  }

  #create $reqmsg->{"dataIDs"} and
  #       $reqmsg->{"metadataIDs"} hashes from document
  my ($errorstring, $metaid) = $reqmsg->parse_all;

  if($errorstring){
    print_log("info",$errorstring);
    $reqmsg->return_result_code("error.common.parse_error", "$errorstring", $metaid);
    return $reqmsg;
  }

  #check if at least one metadata and one data element is in message
  if(!($reqmsg->{"metadataIDs"})){
    $errorstring = "No metadata definition in message.";
    print_log("info",$errorstring);
    $reqmsg->return_result_code("error.common.message", "$errorstring", "message");
    return $reqmsg;
  }

  if(!(defined $reqmsg->{"dataIDs"})){
    $errorstring = "No data trigger in message.";
    print_log("info",$errorstring);
    $reqmsg->return_result_code("error.common.message", "$errorstring", "message");
    return $reqmsg;
  }
  
  #do some checks on metadata content
  foreach my $meta (keys %{$reqmsg->{"metadataIDs"}}){
    #check for unknown eventTypes
    my $et = $reqmsg->{"metadataIDs"}{$meta}{"eventType"};
    if ($et =~ /admin/){ #dispatch to Echo module
      return perfSONAR::Echo::handle_echo_request($reqmsg, $service);
      return $reqmsg;
    }

    my $found = undef;
    #foreach my $sup_et (@{$services{$service}{"eventtype"}}){
    foreach my $sup_et (@module_ets){
      next unless ($sup_et =~ /$et/);
      $found = 1;
    }
    if (!defined $found){
      my $errorstring = "Unknown eventType: $et";
      print_log("info",$errorstring);
      $reqmsg->return_result_code("error.common.parse_error", $errorstring, $meta);
      return $reqmsg;
    }
    #check times
    my $startTime = $reqmsg->{"metadataIDs"}{$meta}{"startTime"};
    my $endTime = $reqmsg->{"metadataIDs"}{$meta}{"endTime"};
    if ($endTime && $startTime && ($endTime < $startTime)){
      my $errorstring = "Illegal time duration specified: " .
        "$endTime is later than $startTime!";
      print_log("info",$errorstring);
      $reqmsg->return_result_code("error.common.parse_error", $errorstring, $meta);
      return $reqmsg;
    }
  }

  #add metadata parameters to data hashes
  ($errorstring, $metaid) = $reqmsg->concatenate_params;
  if ($errorstring){
    print_log("info", $errorstring);
    $reqmsg->return_result_code("error.common.parse_error", "$errorstring", $metaid);
    return $reqmsg;
  }

  #DEBUG output:
  print Dumper ($reqmsg->{"dataIDs"});
  #print Dumper ($reqmsg->{"metadataIDs"});


  #initialize data module
  $services{$service}->{handler}->open_request();

  #start action for each data block
  foreach my $dataid (keys %{$reqmsg->{"dataIDs"}}){
    $services{$service}->{handler}->start_action($reqmsg, $dataid, $service);
  }

  $services{$service}->{handler}->close_request();

  return $reqmsg;
}


sub sendReceive {
  my %p = (
    message => undef,
    host => "localhost",
    port => "8090",
    endpoint => "/",
    uri => "",
    soapheader => "", #TODO TODO noch nicht fertig und kein String!
    @_
  );

  my $uri = $p{uri} || "http://$p{host}:$p{port}$p{endpoint}";
  my $body = $p{message}->as_dom; 
 
  # Message is NMWG::Message
  my $message = perfSONAR::SOAP::Message->new(
    body => $body,
    uri => $uri
  );
  # Modify SOAP header using $message or set via "new" call directly

  my $userAgent = perfSONAR::SOAP::HTTP::UserAgent->new;
  my $request = perfSONAR::SOAP::HTTP::Request->new(message => $message);

  my $response = $userAgent->request($request);
  unless ($response->is_success) {
    # HTTP error
    #my $code = $response->code();
    #my $message = $response->message();
    carp $response->status_line if $^W; # "<code> <message>"
    return;
  }
 
  my $soap_message = $response->soap_message;
  if ($soap_message->is_fault) {
    carp "TODO" if $^W;
    return;
  }

  my $nmwg_message = NMWG::Message->new( ($soap_message->body)[0] );
  #TODO This should perhaps be NMWG::Message->from_soap_message($soap_message);
  
  return $nmwg_message;
}


sub print_log_old {
  my ($level,$message) = @_;

  if (defined $log) {
    $level = "debug" unless $log->level_is_valid($level);
    $log->log(level => $level, message => $message);
  } else {
    print STDERR "$level: $message\n";
  }
  return 1;
}

sub print_log {
  my ($level,$message,$service) = @_;

  if (defined $log) {
    $level = "debug" unless $log->level_is_valid($level);
    $log->log(level => $level, service => $service, message => $message);
  } else {
    print STDERR ($service ? "$service: " : "unknown service: ") .
      "$level: $message\n";
  }
  return 1;
}
1;
