package perfSONAR::MP;

BEGIN {
  use vars qw(@ISA);
  #@ISA = qw(SOAP::Server::Parameters);
  @ISA = qw(Exporter);
  @EXPORT = qw(store check_parameters get_parameters);
}



#DEBUG
use Data::Dumper;
#/DEBUG

use strict;
use warnings;

use Carp;
use perfSONAR qw(print_log);
use NMWG;
use NMWG::Message;


sub new {
  my $this = shift;
  my $class = ref($this) || $this;
  my %param = (@_); 

  my $self = {
  };
  $self->{"supportedEventtypes"} = [
    "http://ggf.org/ns/nmwg/ops/store/2.0",
  ];
  $self->{"command"} = $param{"command"};
  $self->{"store"} = $param{"store"};
  $self->{"store_url"} = $param{"store_url"};

  bless $self, $class;
  return $self;
}

sub open_request {
  #initialize whatever you have to initiliaze
}

sub close_request{
  #close request if necessary
}


#obsolete!
sub set_command {
  my $self = shift;
  my $c = shift;
  $self->{"command"} = $c;
}

sub selftest {
  my $self = shift;
  return undef;
}


sub start_action {

}


sub store {

  my $self = shift;
  my $msg = shift;

  if (!$self->{"store_url"}){
    $self->print_log_id("error", "Store to MA failed: no url to SQL MA given!");
  } else {
    my $store_msg = $msg->clone;
    if (!$store_msg){
      $self->print_log_id("error", "Store to MA failed: Could not clone storage message.");
      return;
    }
    $store_msg->set_message_type("MeasurementArchiveStoreRequest");

    #my $store_string = $store_msg->as_string(1);
    #print "$store_string\n";

    my $response = perfSONAR::sendReceive( message => $store_msg,
                            uri => $self->{"store_url"},
                          );
    
    if ($response){
      my $result = $response->as_string(2);
      $self->print_log_id("info", "Answer from SQL MA:\n$result"); 
      #print "$result\n"; #TODO parse response?
    } else {
      $self->print_log_id("error", "Store to MA failed: No response from SQL MA!");
    }
  }
}


sub get_parameters {

  my $self = shift;
  my $datablock = shift;

  my %parameters; 

  foreach my $key (keys %{$datablock}){ #get eventtypes
    next if ($key eq "node" || $key eq "metaref" ); #no eventtype
    foreach my $k (keys %{$datablock->{$key}}){
      if (!defined $parameters{$k}){
        if ($k eq "src" || $k eq "dst"){
          $parameters{$k} = $datablock->{$key}{$k}{"value"};
        } else {
          $parameters{$k} = $datablock->{$key}{$k};
        }
      }
    }
  }

  return %parameters;
}


sub check_parameters {
  
  my $self = shift;
  my %params = @_;

  my @unknown;
  my @unsupported;


  foreach my $par (keys %params){
    next if ($par eq "ns_prefix" ||
             $par eq "param_ns_prefix" ||
             $par eq "subject_ns_prefix" ||
             $par eq "subject_ns_uri" ||
             $par eq "param_ns_uri" ||
             $par eq "parameter_ID" ||
             $par eq "metaID" ||
             $par eq "address" ||
             $par eq "metadatakey"
            );
    next if exists $self->{"known_parameters"}{$par};
    if (exists $self->{"unsupported_parameters"}->{$par}){
      push @unsupported, $par;
    } else {
      push @unknown, $par;
    }

  }
#TODO
    if ($#unknown >= 0){
    my $error =  "Unknown parameter(s): " . join (", ", @unknown);
    return $error;
  }

  if ($#unsupported >= 0){
    my $error =  "Unsupported parameters(s): " . join (", ", @unsupported);
    return $error;
  }
}


sub print_log_id {
  my ($self, $level,$message) = @_;
  print_log($level,$message,$self->{"service_id"});
}

1;
