package perfSONAR::MP::BWCTL;

use vars qw(@ISA);
@ISA = qw(perfSONAR::MP);

#DEBUG
use Data::Dumper;
#/DEBUG

use strict;
use warnings;

use Carp;

use perfSONAR::BWCTL;
use perfSONAR::MP;

BEGIN {
    use vars qw($VERSION);
    $VERSION     = 0.51;
}

my $ns_store = "http://ggf.org/ns/nmwg/ops/store/2.0/";

sub new {
  my $this = shift;

  my $self = perfSONAR::MP->new (@_);  
  push @{$self->{"supportedEventtypes"}},

    ("bwctl",
    "http://ggf.org/ns/nmwg/tools/bwctl/2.0/",
    "http://ggf.org/ns/nmwg/tools/iperf/2.0/");
 
  $self->{"known_parameters"} = {

    src => 1,
    dst => 1,
    interval => 1,
    duration => 1,
    windowSize => 1,
    protocol => 1,
    bufferSize => 1,
    bandwidth => 1,
    login => 1,
    password => 1,
    TOS => 1,
  };

  $self->{"unsupported_parameters"} = {

    advisoryWindowsize => 1,
    scheduleInterval => 1,
    numberOfTests => 1,
    latest => 1,
    alpha => 1,
  };

  bless $self, qw(perfSONAR::MP::BWCTL);
  return $self;
}

sub parse_key {

  my $self = shift;
  my %parameters = shift;
  my %key = split (',', $parameters{"metadatakey"});
  $parameters{"src"} = $key{"src"}{"value"} if $key{"src"}{"value"};
  $parameters{"dst"} = $key{"dst"}{"value"} if $key{"dst"}{"value"};
  $parameters{"interval"} = $key{"int"} if $key{"int"};
  $parameters{"duration"} = $key{"dur"} if $key{"dur"};
  $parameters{"windowSize"} = $key{"win"} if $key{"win"};
  $parameters{"protocol"} = $key{"pro"} if $key{"pro"};
  $parameters{"bufferSize"} = $key{"buf"} if $key{"buf"};
  $parameters{"bandwidth"} = $key{"ban"} if $key{"ban"};
  $parameters{"TOS"} = $key{"tos"} if $key{"tos"};
}


sub selftest {
  my $self = shift;
  return {
    bwctl_command_test => \&bwctl_command_test, 
    bwctl_exec_test => \&bwctl_exec_test, 
    bwctld_running_test => \&bwctld_running_test, 
    ntpd_running_test => \&ntpd_running_test,
  };
}


sub bwctl_command_test{
  my $self = shift;
  my $message;    
  my $status = "error";
  my $command = $self->{"command"};
  my $commandpath = `which $command`;
  chomp ($commandpath);
  if (!(-e "$commandpath")){
    $message = "BWCTL tool $command not found!";
    #die "BWCTL tool $commandpath not found!\n";
  } else {
    $message = "BWCTL tool $command found ";
    $status = "success";
  }
  return ($message, $status);
}

sub bwctl_exec_test{
  my $self = shift;
  my $message;    
  my $status = "error";
  my $command = $self->{"command"};
  my $commandpath = `which $command`;
  chomp ($commandpath);
  if (!( -x "$commandpath")){
    $message = "BWCTL tool $command not executable!";
    #die "No executable $commandpath found!\n";
  } else {
    $message = "BWCTL tool $command executable.";
    $status = "success";
  }
  return ($message, $status);
}

sub bwctld_running_test{
  my $self = shift;
  my $message;
  my $status = "error";
  my @ps_output = `ps auxw |grep bwctld`;
  my $bwctld = undef;
  while (my $elem = pop @ps_output){
    next if $elem =~ /grep/;
    $bwctld = 1;
  }
  if ($bwctld){
    $message = "bwctld running.";
    $status = "success";
  } else {
    $message = "bwctld not running! ";
    $self->print_log_id ("warning", "bwctld not running!");
  }
  return ($message, $status);
}

sub ntpd_running_test{
  my $self = shift;
  my $message;
  my $status = "error";
  my @ps_output = `ps auxw |grep ntpd`;
  my $ntpd = undef;
  while (my $elem = pop @ps_output){
    next if $elem =~ /grep/;
    $ntpd = 1;
  }
  if ($ntpd){
    $message = "ntpd running.";
    $status = "success";
  } else {
    $message .= "ntpd not running!";
    $self->print_log_id ("warning", "ntpd not running!");
  }
  return ($message, $status);
}


sub start_action {
  my $self = shift;
  my $msg = shift;
  my $dataid = shift;
  my $service_id = shift;

  $self->{"service_id"} = shift;

  my %parameters;
  my $ns = "http://ggf.org/ns/nmwg/tools/iperf/2.0/";

  %parameters = $self->get_parameters($msg->{"dataIDs"}{$dataid});

  my $error = $self->check_parameters(%parameters);

  if ($error){
    $self->print_log_id ("info", $error);
    return $msg->return_result_code("error.mp.bwctl", $error, $msg->{"dataIDs"}{$dataid}{"metaref"});
  }  

  my @datalines;

  my $messagetype = $msg->get_message_type();
  
  if ($messagetype =~ /EchoResponse/){
    #do selftests
    my @tests = ("bwctl_command_test",
                 "bwctl_exec_test",
                 "bwctld_running_test",
                 "ntpd_running_test");
    foreach my $test (@tests){
      $self->print_log_id("debug", "Perform test: $test");
      my ($message, $status) = $self->$test;
      my $et = "http://schemas.perfsonar.net/tools/admin/selftest/MP/BWCTL/$test/$status/1.0";
      $msg->return_result_code(
        $et, $message, $msg->{"dataIDs"}{$dataid}{"metaref"}, $test);
    }
    return 1;
  }
  
  if ($messagetype eq "MetadataKeyResponse") {
    my %data_hash;
    my $key;
    $key .= "src,$parameters{src}" if $parameters{"src"};
    $key .= ",dst,$parameters{dst}" if $parameters{"dst"};
    $key .= ",int,$parameters{interval}" if $parameters{"interval"};
    $key .= ",dur,$parameters{duration}" if $parameters{"duration"};
    $key .= ",win,$parameters{windowSize}" if $parameters{"windowSize"};
    $key .= ",pro,$parameters{protocol}" if $parameters{"protocol"};
    $key .= ",buf,$parameters{bufferSize}" if $parameters{"bufferSize"};
    $key .= ",ban,$parameters{bandwidth}" if $parameters{"bandwidth"};
    $key .= ",tos,$parameters{TOS}" if $parameters{"TOS"};

    $data_hash{"MetadataKey"} = "$key";
    push @datalines, \%data_hash;
    $msg->set_data($dataid, @datalines);
    return 1;
  }

  my $unixtime = time;
  my @commandline = perfSONAR::BWCTL::commandline(%parameters);
  if ($commandline[0] eq "ERROR") {
    return $msg->return_result_code("error.mp.bwctl", $commandline[1], $msg->{"dataIDs"}{$dataid}{"metaref"});
  }
  $self->print_log_id("info", "BWCTL commandline: @commandline");
  my @result = perfSONAR::BWCTL::do_bwctl($self->{"command"}, $parameters{"password"}, @commandline);
  @datalines = perfSONAR::BWCTL::parse_result($unixtime, @result);
  if ($datalines[0] eq "BWCTL Error:"){
    my $errorstring = join(" ", @datalines);
    $errorstring =~ s/usage.*$//;
    return $msg->return_result_code("error.mp.bwctl", $errorstring, $msg->{"dataIDs"}{$dataid}{"metaref"});
  }

  $msg->set_data_ns ($dataid, $ns, @datalines);

=cut
  my $datanode = $msg->{"dataIDs"}{$dataid}{"node"};
  my $timenode = $msg->add_attribute(parent => $datanode, nodename => "commonTime",
                           type => "unix", value => $unixtime);
  foreach my $data_hash (@datalines){

    $msg->add_attribute(parent => $timenode, nodename => "datum",
                  namespace => $ns, %{$data_hash});
  }
=cut

  if ($self->{"store"} == 1){
    $self->store($msg);
  }


  return 1;

}

1;
